@@ -166,7 +166,7 @@ require Exporter;
 @EXPORT = qw (
   $Script $BasePath $LibPath $BinPath $CfgPath $ImagePath $DataPath
   $TmpBase %ConfigData ReadFile RealRPM RealRPMs ReadRPM $SUBinary SUSystem Print2File $MToolsCfg $AutoBuild
-  ResolveDeps
+  ResolveDeps CopyRPM
 );
 
 use strict 'vars';
@@ -387,6 +387,23 @@ sub UnpackRPM
 }
 
 
+#
+# Unpack rpm_name to target_dir and return path to cache dir or undef if failed.
+#
+sub CopyRPM
+{
+  my $rpm_name = $_[0];
+  my $target_dir = $_[1];
+
+  my $cache_dir = ReadRPM($rpm_name);
+  if($cache_dir) {
+    system "tar -C '$cache_dir/rpm' -cf - . | tar -C '$target_dir' -xf -";
+  }
+
+  return $cache_dir;
+}
+
+
 #
 # Unpack rpm to cache dir and return path to dir or undef if failed.
 #
@@ -464,20 +481,35 @@ sub ReadRPM
       undef $kv;
     }
 
-    UnpackRPM RealRPM("$rpm->{name}-base"), $tdir;
-    UnpackRPM RealRPM("$rpm->{name}-extra"), $tdir;
-    UnpackRPM RealRPM("$rpm->{name}-optional"), $tdir;
+    CopyRPM "$rpm->{name}-base", $tdir;
+    CopyRPM "$rpm->{name}-extra", $tdir;
+    CopyRPM "$rpm->{name}-optional", $tdir;
 
     my $kmp;
     for (split(',', $ConfigData{kmp_list})) {
       ($kmp = $rpm->{name}) =~ s/^kernel/$_-kmp/;
       print "adding kmp $kmp\n";
-      UnpackRPM RealRPM($kmp), $tdir;
+      CopyRPM $kmp, $tdir;
     }
 
     for (split(',', $ConfigData{fw_list})) {
-      print "adding firmware $_\n";
-      UnpackRPM RealRPM($_), $tdir;
+      if($_ eq 'kernel-firmware-all') {
+        my $rpm = ReadRPM($_);
+        if($rpm) {
+          my $deps = ReadFile "$rpm/requires";
+          for my $fw (split /\n/, $deps) {
+            if($fw =~ /^(kernel-firmware-\S+)/) {
+              $fw = $1;
+              print "adding firmware $fw\n";
+              CopyRPM $fw, $tdir;
+            }
+          }
+        }
+      }
+      else {
+        print "adding firmware $_\n";
+        CopyRPM $_, $tdir;
+      }
     }
 
     # keep it readable
@@ -490,6 +522,41 @@ sub ReadRPM
       print "warning: kmp/firmware version mismatch: $_\n";
       SUSystem "sh -c 'tar -C $tdir/lib/modules/$_ -cf - . | tar -C $tdir/lib/modules/$kv -xf -'";
     }
+
+    # unpack kernel modules if requested
+    # see doc/compression.md
+    if($ENV{instsys_no_compression} =~ /modules/) {
+      my $dir = "$tdir/usr/lib/modules";
+      $dir = "$tdir/lib/modules" if ! -d $dir;
+
+      print "uncompressing kernel modules...\n";
+
+      SUSystem "find $dir -type f -name \*.ko.zst -exec zstd -d --quiet --rm '{}' \\;";
+      SUSystem "find $dir -type f -name \*.ko.xz -exec xz -d '{}' \\;";
+      SUSystem "find $dir -type f -name \*.ko.gz -exec gzip -d '{}' \\;";
+    }
+
+    # unpack kernel firmware if requested
+    # see doc/compression.md
+    if($ENV{instsys_no_compression} =~ /firmware/) {
+      my $dir = "$tdir/usr/lib/firmware";
+      $dir = "$tdir/lib/firmware" if ! -d $dir;
+
+      print "uncompressing kernel firmware...\n";
+
+      # rename symlinks
+      for my $suffix ("zst", "xz", "gz") {
+        SUSystem "find $dir -type l -name \*.$suffix -exec rename -sl .$suffix '' '{}' \\; -exec rename -l .$suffix '' '{}' \\;";
+      }
+
+      # unpack files
+      SUSystem "find $dir -type f -name \*.zst -exec zstd -d --quiet --rm '{}' \\;";
+      SUSystem "find $dir -type f -name \*.xz -exec xz -d '{}' \\;";
+      SUSystem "find $dir -type f -name \*.gz -exec gzip -d '{}' \\;";
+
+      my $broken = `find $dir -follow -type l`;
+      die "firmware uncompressing left broken symlinks:\n$broken\n" if $broken ne "";
+    }
   }
 
   return $err ? undef : $dir;
@@ -1215,6 +1282,8 @@ $ConfigData{fw_list} = $ConfigData{ini}{Firmware}{$arch} if $ConfigData{ini}{Fir
   $in_abuild = $ConfigData{buildenv}{BUILD_BASENAME} ? 1 : 0;
   $in_abuild = 1 if -d "$ConfigData{buildroot}/.build.binaries";
 
+  $ConfigData{in_abuild} = $in_abuild;
+
   # print STDERR "abuild = $in_abuild\n";
 
   die "\nError: *** you must be root to build images ***\n\n" if $>;
